using System;
using System.Net;
using System.IO;
using System.CodeDom;
using System.CodeDom.Compiler;
using System.Collections;
using System.ComponentModel;
using System.Web.Services;
using System.Web.Services.Description;
using System.Reflection;
using Microsoft.CSharp;
using System.Collections.Specialized;
using System.Text;
using System.Xml.Serialization;

namespace DynamicWebService
{
  /// <summary>
  /// Summary description for DynamicService.
  /// </summary>
  public class DynamicService
  {
    string sProtocol = "Soap"; // Soap, HttpGet, HttpPost
    RequestParameters Req = null;

    public string ExecuteMethod(RequestParameters Request) 
    {
      Req = Request;
      Assembly dynAssem = null;		// assembly for the dynamic web service proxy

      try 
      {
        //load existing WebService dll or create the dll from the wsdl file
        if ( Req.WsdlFile == "" )
          dynAssem = Assembly.LoadFrom(Req.DllName);
        else
          dynAssem = CreateAssembly();
       
        //get the type of the assembly
        Type dynType = dynAssem.GetType(Req.dynNameSpace + "." + Req.ServiceName);
        
        // create an instance of the type
        Object wsInstance = Activator.CreateInstance(dynType);

        //set credentials 
        if ( Req.WsdlFile == "" && Req.User != "" )
        {
          PropertyInfo piCred = dynType.GetProperty("Credentials");
          piCred.SetValue(wsInstance, GetCredentials(), null);
        }

			  //get the method we will call
        MethodInfo mi = wsInstance.GetType().GetMethod(Req.MethodName);
        // call the method
        object result = mi.Invoke(wsInstance, CreateParams(mi.GetParameters()));

        string sReturn = "";
        if ( result.GetType().IsClass )
          //serialize class to xml string
          sReturn = SerializeResult(result);
        else
          sReturn = result.ToString();

        return sReturn;
      }
      catch (Exception e) 
      {
        return e.ToString();
      }
    }

    private NetworkCredential GetCredentials()
    {
      return new NetworkCredential(Req.User, Req.Password, Req.Domain);
    }

    private string SerializeResult(object Returned)
    {
      XmlSerializer serializer = new XmlSerializer(Returned.GetType().UnderlyingSystemType);
      MemoryStream ms = new MemoryStream();
      serializer.Serialize(ms, Returned);

      ms.Seek(0, SeekOrigin.Begin);

      byte[] temp = new byte[ms.Length];
      int i = ms.Read(temp, 0, (int)ms.Length);
      ASCIIEncoding AE = new ASCIIEncoding();

      return AE.GetString(temp, 0, temp.Length);
    }

    private object[] CreateParams(ParameterInfo[] pi)
    {
      // Create the args array
      Object[] args = new Object[pi.GetUpperBound(0)+1];
      
      //get each req'd param and fill in the value
      for ( int idx = 0; idx <= pi.GetUpperBound(0); idx++)
        //convert values to parameter types
        args[pi[idx].Position] = Convert.ChangeType(Req.getParameter(idx), pi[idx].ParameterType);
      
      return args;
    }

    private Assembly CreateAssembly()
    {
      ServiceDescriptionImporter sdImp = new ServiceDescriptionImporter();

      // get the wsdl file info
      ServiceDescription sDesc = ServiceDescription.Read(GetWsdl());

      sdImp.ProtocolName = sProtocol;
      sdImp.AddServiceDescription(sDesc, null, null);
			
      CodeNamespace cnSpace = new CodeNamespace(Req.dynNameSpace);
      CodeCompileUnit ccUnit = new CodeCompileUnit();

      ServiceDescriptionImportWarnings sdiWarning = sdImp.Import(cnSpace, ccUnit);
      if ( sdiWarning == ServiceDescriptionImportWarnings.NoCodeGenerated 
        || sdiWarning == ServiceDescriptionImportWarnings.NoMethodsGenerated )
        throw new Exception(string.Format("Import failed with {0} return code.", sdiWarning)); 

      // Pass the CodeCompileUnit to a System.CodeDom.CodeProvider to do the compilation. 
      CSharpCodeProvider cdp = new CSharpCodeProvider();
			
      // source code generation
      ICodeGenerator cg = cdp.CreateGenerator();

      // this is for source code
      StringBuilder sb = new StringBuilder();
      StringWriter sw = new StringWriter(sb);

      // generate the WS proxy code.
      cg.GenerateCodeFromNamespace(cnSpace, sw, null);
      string fileInfo = sb.ToString();

      sw.Close();

      CompilerParameters cp = new CompilerParameters();
      cp.ReferencedAssemblies.Add("System.dll");
      cp.ReferencedAssemblies.Add("System.Xml.dll");
      cp.ReferencedAssemblies.Add("System.Web.Services.dll");
      cp.ReferencedAssemblies.Add("System.Data.dll");
      cp.GenerateExecutable = false;
      
      cp.IncludeDebugInformation = false;

      // we MUST save the assembly to a dll file in order for the XmlSerializer to function
      cp.GenerateInMemory = false;
      cp.OutputAssembly = Req.DllName;

      //compile code
      ICodeCompiler icc = cdp.CreateCompiler();
      CompilerResults cr = icc.CompileAssemblyFromSource(cp, fileInfo);

      if(cr.Errors.HasErrors)
      {
        StringBuilder se = new StringBuilder();
        for ( int i = 0; i < cr.Errors.Count; i++ )
        {
          se.Append(cr.Errors[i].ErrorText);
          se.Append("  ");
        }
        throw new Exception(se.ToString()); 
      }

      //return the compiled assembly
      return cr.CompiledAssembly;
      
    }

    public string Protocol
    {
      get { return sProtocol; }
      set { sProtocol = value; }
    }

    private Stream GetWsdl()
    {
      // this could be a valid WSDL representation
      if(Req.WsdlFile.StartsWith("<?xml version") == true)
      {
        ASCIIEncoding AE = new ASCIIEncoding();
        byte[] bytes = AE.GetBytes(Req.WsdlFile);
        MemoryStream ms = new MemoryStream(bytes);
        return (Stream)ms;						
      }
      // this is a URL to the WSDL
      else if(Req.WsdlFile.StartsWith("http://") == true)
      {
        return GetWsdlFromUri();			
      }
					
      // try to get from local file system
      return GetWsdlFromFile();			
    }
		
    private Stream GetWsdlFromUri()
    {
      WebRequest req = WebRequest.Create(Req.WsdlFile);

      if ( Req.Proxy != "" )
      {
        WebProxy myProxy = new WebProxy(Req.Proxy, Req.Port);
        myProxy.BypassProxyOnLocal = true;

        req.Proxy = WebProxy.GetDefaultProxy();
      }

      req.Credentials = GetCredentials();
      
      Stream objStream;
      objStream = req.GetResponse().GetResponseStream();
      return objStream;
    }
		
    private Stream GetWsdlFromFile()
    {
      FileInfo fi = new FileInfo(Req.WsdlFile);
			
      if(fi.Extension == "wsdl")
      {
        FileStream fs = new FileStream(Req.WsdlFile, FileMode.Open, FileAccess.Read);
        return (Stream)fs;
      }
				
      throw new Exception("This is not a WSDL file");
    }

  }

  public class RequestParameters 
  {
    System.Collections.ArrayList paramValues = new System.Collections.ArrayList();
    string swsdlfile = "", sdynNameSpace = "";
    string sDllName = "", sServiceName = "", sMethodName="";
    string sProxy = "", sUser = "", sPassword = "", sDomain = "";
    int iPort = 0;

    //credentials
    public string Domain
    {
      get { return sDomain; }
      set { sDomain = value; }
    }

    //credentials
    public string Password
    {
      get { return sPassword; }
      set { sPassword = value; }
    }

    //credentials
    public string User
    {
      get { return sUser; }
      set { sUser = value; }
    }

    //Proxy address
    public string Proxy
    {
      get { return sProxy; }
      set { sProxy = value; }
    }

    //Proxy port
    public int Port
    {
      get { return iPort; }
      set { iPort = value; }
    }

    //Save or load dll name
    public string DllName
    {
      get { return sDllName; }
      set { sDllName = value; }
    }

    //One of:
    // url to wsdl file
    // path to local wsdl file
    // actual wsdl file text
    public string WsdlFile 
    {
      get { return swsdlfile; }
      set { swsdlfile=value; }
    }
    
    //the namespace of our class
    public string dynNameSpace 
    {
      get { return sdynNameSpace; }
      set { sdynNameSpace=value; }
    }
    
    //WebService service name
    public string ServiceName 
    {
      get { return sServiceName; }
      set { sServiceName=value; }
    }

    //The method we want to call in the web service
    public string MethodName 
    {
      get { return sMethodName; }
      set { sMethodName=value; }
    }
    
    // add a parameter to the web service call
    public void addParameter(string paramvalue) 
    {
      paramValues.Add(paramvalue);
    }
    
    // return count of paramters
    public int count() 
    {
      return paramValues.Count;
    }
 
    //get the parameter value
    public string getParameter(int whichone)
    {
      return Convert.ToString(paramValues[whichone]);
    }
  }
}
